from pydantic import BaseModel


class BaseRequest(BaseModel):
    """Base class for API requests."""

    @classmethod
    def example(cls):
        field_values = {}
        for name, field in cls.model_fields.items():
            if field.examples is not None and len(field.examples) > 0:
                field_values[name] = field.examples[0]
            else:
                field_values[name] = field.default
        return cls(**field_values)



import asyncio
import base64
import json
import logging
import os
import urllib
import urllib.parse
import urllib.request
from typing import Any, Dict, List, Optional, Union
from urllib import request

import httpx
import requests

from creator.base.base_app import BaseApp
from creator.base.base_response import ResponseDataType, ResponseData, BaseResponse
from creator.base.job_status import JobStatus
from creator.cmfy.request_cmfy import CmfyWorkflow, CmfyWorkflowFcus, model_load_json
from creator.cmfy.response_cmfy import ResponseCmfy

logger = logging.getLogger(__name__)


class AppCmfy(BaseApp):
    param_classes = [CmfyWorkflow, CmfyWorkflowFcus]
    output = {}

    def __init__(self):
        super().__init__()
        focus_port = os.environ.get("PORT_CMFY", 8888)
        if isinstance(focus_port, str):
            focus_port = int(focus_port)
        self.api_base_url = f"http://0.0.0.0:{focus_port}"

    async def create(self, params: Union[CmfyWorkflow, CmfyWorkflowFcus]) -> ResponseCmfy:
        cmfy_port = os.environ.get("PORT_CMFY", 8889)
        url = f"http://localhost:{cmfy_port}/prompt"
        workflow = json.loads(params.workflow_json)
        self._check_loaded_model(workflow)
        logger.debug(f"CMFY data: {workflow}")
        # Send a POST to cmfy_url with the workflow json
        # The response will be JSON string with the below format
        response_json = {}
        try:
            headers = {'Content-Type': 'application/json'}
            data = {"prompt": workflow}
            data = json.dumps(data).encode('utf-8')
            response = requests.post(url, data=data, headers=headers)
            logger.debug(f"CMFY response: {response}, {response.text}, {response.json()}")
            # If we have a 422 here, print what the unprocessable entity is
            if response.status_code == 422:
                logger.warning(f"Unprocessable entity: {response.text}")
            response.raise_for_status()
            response_json = response.json()
        except Exception as e:
            logger.error(f"Error creating prompt: {e}")
            return ResponseCmfy.error(f"Error creating prompt: {e} ({response_json})")
        # {"prompt_id": "f57c69ee-f4ed-4866-b5a4-4072c17c36a8", "number": 2, "node_errors": {}}
        prompt_id = response_json.get("prompt_id", None)
        logger.debug(f"Prompt ID: {prompt_id}")
        if prompt_id is None:
            return ResponseCmfy.error(f"Error creating prompt: {response_json}")
        status_count = 0
        status = await self.get_status(prompt_id)
        while status != JobStatus.FINISHED and status != JobStatus.FAILED and status_count < 10:
            await asyncio.sleep(1)
            status = await self.get_status(prompt_id)
            status_count += 1
        if status == JobStatus.FINISHED:
            data = self.output
            response_data = ResponseData(
                data=data,
                data_type=ResponseDataType.IMAGE,
                total_count=len(data)
            )
            logger.info(f"Success, returning response with {len(data)} images")
            return ResponseCmfy.success(response_data, prompt_id)
        return ResponseCmfy.error("Error creating prompt")

    async def create_async(self, params: Union[CmfyWorkflow, CmfyWorkflowFcus]) -> ResponseCmfy:
        cmfy_port = os.environ.get("PORT_CMFY", 8889)
        url = f"http://localhost:{cmfy_port}/prompt"
        workflow = json.loads(params.workflow_json)
        self._check_loaded_model(workflow)
        logger.debug(f"CMFY data: {workflow}")
        # Send a POST to cmfy_url with the workflow json
        # The response will be JSON string with the below format
        response_json = {}
        try:
            headers = {'Content-Type': 'application/json'}
            data = {"prompt": workflow}
            data = json.dumps(data).encode('utf-8')
            response = requests.post(url, data=data, headers=headers)
            logger.debug(f"CMFY response: {response}, {response.text}, {response.json()}")
            # If we have a 422 here, print what the unprocessable entity is
            if response.status_code == 422:
                logger.warning(f"Unprocessable entity: {response.text}")
            response.raise_for_status()
            response_json = response.json()
        except Exception as e:
            logger.error(f"Error creating prompt: {e}")
            return ResponseCmfy.error(f"Error creating prompt: {e} ({response_json})")
        # {"prompt_id": "f57c69ee-f4ed-4866-b5a4-4072c17c36a8", "number": 2, "node_errors": {}}
        prompt_id = response_json.get("prompt_id", None)
        logger.debug(f"Prompt ID: {prompt_id}")
        if prompt_id is None:
            return ResponseCmfy.error(f"Error creating prompt: {response_json}")
        return ResponseCmfy.running(prompt_id)

    async def get_status(self, job_id) -> BaseResponse:
        url = f"http://localhost:{os.environ.get('PORT_CMFY', 8889)}/history/{job_id}"
        job_data = None
        print(f"Getting status for job {job_id} from url {url}")
        logger.debug(f"Getting status for job {job_id}")
        while job_data is None:
            history = requests.get(url)
            if history.status_code != 200:
                return BaseResponse.error("Error getting job status", job_id)
            logger.info(f"History: {history.json()}")
            history = history.json()
            job = history.get(job_id, None)
            if not job:
                continue
            job_data = job

        status = job_data.get('status', {})
        status_str = status.get('status_str', None)
        completed = status.get('completed', False)
        messages = status.get('messages', [])
        if status_str == "success" and completed:
            images_output = []
            videos_output = []
            for node_id, node_output in job_data['outputs'].items():
                # TODO: Add handler here for animated images, etc.
                if 'images' in node_output:
                    logger.debug(f"Node output: {node_output}")
                    for image in node_output['images']:
                        logger.debug(f"Getting image: {image['filename']}")
                        print(f"Getting image: {image['filename']}")
                        image_data = await self._get_image(image['filename'], image['subfolder'], image['type'])
                        images_output.append(image_data)
                    print(f"We are returning {len(images_output)} images")
                    response_data = ResponseData(data=images_output, data_type=ResponseDataType.IMAGE,
                                                 total_count=len(images_output))
                    return ResponseCmfy.success(data=response_data)
                if 'gifs' in node_output:
                    output_is_gif = False
                    print(f"Node output: {node_output}")
                    for gif in node_output['gifs']:
                        print(f"Getting gif: {gif['filename']}")
                        gif_data = await self._get_image(gif['filename'], gif['subfolder'], gif['type'])
                        output_is_gif = gif['filename'].endswith(".gif")
                        videos_output.append(gif_data)
                    print(f"We are returning {len(videos_output)} gifs/videos")
                    response_type = ResponseDataType.GIF if output_is_gif else ResponseDataType.VIDEO
                    response_data = ResponseData(data=videos_output, data_type=response_type,
                                                 total_count=len(videos_output))
                    return ResponseCmfy.success(data=response_data)
        if status_str == "error":
            error_message = f"An exception occurred: {messages}"
            return ResponseCmfy.error(error_message, job_id)
        if not completed:
            return ResponseCmfy.running(job_id)
        return ResponseCmfy.error("Error getting job status", job_id)

    async def get_models(self) -> List[Any]:
        url = f"http://localhost:{os.environ.get('PORT_CMFY', 8889)}/object_info"
        async with httpx.AsyncClient() as client:
            response = await client.get(url)
            response.raise_for_status()
            response_json = response.json()
            models = []
            loader = response_json.get("CheckpointLoaderSimple", None)
            if loader:
                input = loader.get("input", None)
                if input:
                    required = input.get("required", None)
                    if required:
                        model_name = required.get("ckpt_name", None)
                        if model_name:
                            models.append(model_name)
        return models

    async def upload_image(self, image_data: Any) -> str:
        pass

    async def test(self):
        url = f"http://localhost:{os.environ.get('PORT_CMFY', 8889)}/history"
        async with httpx.AsyncClient() as client:
            response = await client.get(url)
            if response.status_code == 200:
                return BaseResponse.active()
            else:
                return BaseResponse.error("Error testing connection")

    @staticmethod
    async def _get_image(filename: str, subfolder: str, folder_type: str) -> Optional[str]:
        server_address = f"localhost:{os.environ.get('PORT_CMFY', 8889)}"
        data = {"filename": filename, "subfolder": subfolder, "type": folder_type}
        url_values = urllib.parse.urlencode(data)
        req_url = f"http://{server_address}/view?{url_values}"
        logger.debug(f"Getting image from {req_url}")
        try:
            async with httpx.AsyncClient() as client:
                response = await client.get(req_url)
                response.raise_for_status()
                image_data = response.content
                # Convert the image data to a base64 string
                base64_encoded = base64.b64encode(image_data).decode('utf-8')
                print(f"Image data: {base64_encoded[:100]}")
                return base64_encoded
        except Exception as e:
            logger.error(f"Failed to retrieve or encode image: {e}")
            return None

    def _check_loaded_model(self, model_name):
        if model_name and model_name not in self.loaded_models:
            self.loaded_models.append(model_name)

        max_cached_models = os.environ.get("MAX_CACHED_MODELS", 3)
        if len(self.loaded_models) > max_cached_models:
            self.loaded_models = self.loaded_models[-max_cached_models:]

    async def cache_model(self, model: str = "Juggernaut-XI-Prototype.safetensors") -> bool:
        request_json = model_load_json
        request_json["1"]['inputs']['ckpt_name'] = model
        workflow_json_str = json.dumps(request_json)
        data = {"workflow_json": workflow_json_str}
        wf = CmfyWorkflow(**data)
        wf.workflow_json = workflow_json_str
        # Print the workflow
        print(f"Workflow: {wf.workflow_json} {wf.__dict__}")
        self._check_loaded_model(model)
        response = await self.create(wf)
        if response.status == JobStatus.FINISHED:
            return True
        return False




from dataclasses import dataclass, field
from typing import List, Dict, Optional

from pydantic import BaseModel, Field

from creator.base.base_request import BaseRequest


@dataclass
class FcusAdvancedParams:
    adm_scaler_positive: float = 1.5
    adm_scaler_negative: float = 0.8
    adm_scaler_end: float = 0.3
    refiner_swap_method: str = "joint"
    adaptive_cfg: int = 7
    sampler_name: str = "dpmpp_2m_sde_gpu"
    scheduler_name: str = "karras"
    overwrite_step: int = -1
    overwrite_switch: int = -1
    overwrite_width: int = -1
    overwrite_height: int = -1
    overwrite_vary_strength: int = -1
    overwrite_upscale_strength: int = -1
    mixing_image_prompt_and_vary_upscale: bool = False
    mixing_image_prompt_and_inpaint: bool = False
    debugging_cn_preprocessor: bool = False
    skipping_cn_preprocessor: bool = False
    controlnet_softness: float = 0.25
    canny_low_threshold: int = 64
    canny_high_threshold: int = 128
    inpaint_engine: str = "v1"
    freeu_enabled: bool = False
    freeu_b1: float = 1.01
    freeu_b2: float = 1.02
    freeu_s1: float = 0.99
    freeu_s2: float = 0.95


@dataclass
class Lora:
    model_name: str = ""
    weight: float = 1.0


class FcusTxt2Img(BaseRequest):
    prompt: str = ""
    negative_prompt: str = ""
    style_selections: List[str] = Field(default_factory=list, examples=[["Fooocus V2", "Fooocus Enhance", "Fooocus Sharp"]])
    performance_selection: str = "Speed"
    aspect_ratios_selection: str = "1152*896"
    image_number: int = 2
    image_seed: int = -1
    sharpness: float = 0.0
    guidance_scale: float = 7.5
    base_model_name: str = "Juggernaut-XI-Prototype.safetensors"
    refiner_model_name: str = "None"
    refiner_switch: float = 0.1
    loras: List[Lora] = Field(default_factory=list, examples=[[Lora()]])
    advanced_params: Optional[FcusAdvancedParams] = Field(default=None, examples=[FcusAdvancedParams()])
    require_base64: bool = True
    async_process: bool = True


class FcusUpscaleOrVary(BaseRequest):
    input_image: str = ""
    uov_method: str = ""
    prompt: str = ""
    negative_prompt: str = ""
    style_selections: List[str] = Field(default_factory=list,
                                        examples=[["Fooocus V2", "Fooocus Enhance", "Fooocus Sharp"]])
    performance_selection: str = ""
    aspect_ratios_selection: str = ""
    image_number: int = 1
    image_seed: int = -1
    sharpness: float = 0.0
    guidance_scale: float = 7.5
    base_model_name: str = "RunDiffusion-XL-Photo.safetensors"
    refiner_model_name: str = "None"
    refiner_switch: float = 0.1
    loras: List[Lora] = Field(default_factory=list, examples=[[Lora()]])
    advanced_params: Optional[FcusAdvancedParams] = Field(default=None, examples=[FcusAdvancedParams()])
    require_base64: bool = False
    async_process: bool = True


class FcusInpaintOrOutpaint(BaseRequest):
    request_params: str = ""
    input_image: str = ""
    input_mask: str = ""
    outpaint_selections: List[str] = Field(default_factory=list, examples=[])
    prompt: str = ""
    negative_prompt: str = ""
    style_selections: List[str] = Field(default_factory=list,
                                        examples=[["Fooocus V2", "Fooocus Enhance", "Fooocus Sharp"]])
    performance_selection: str = ""
    aspect_ratios_selection: str = ""
    image_number: int = 1
    image_seed: int = -1
    sharpness: float = 0.0
    guidance_scale: float = 7.5
    base_model_name: str = "RunDiffusion-XL-Photo.safetensors"
    refiner_model_name: str = "None"
    refiner_switch: float = 0.1
    loras: List[Lora] = Field(default_factory=list, examples=[[Lora()]])
    advanced_params: Optional[FcusAdvancedParams] = Field(default=None, examples=[FcusAdvancedParams()])
    require_base64: bool = False
    async_process: bool = True



from typing import Dict, Any
from dataclasses import dataclass
from creator.base.base_response import BaseResponse, ResponseData, ResponseDataType


@dataclass
class FcusJobResult:
    base64: str
    url: str
    seed: int
    finish_reason: str


@dataclass
class FcusResult:
    job_id: str
    job_type: str
    job_stage: str
    job_progress: int
    job_status: str
    job_step_preview: str
    job_result: FcusJobResult


class ResponseFcusApi(BaseResponse):
    @classmethod
    def parse_response(cls, response: Dict[str, Any]):
        try:
            job_data = FcusResult(**response)
            if job_data.job_result.base64:
                response_data = ResponseData(data=[job_data.job_result.base64], data_type=ResponseDataType.IMAGE,
                                             total_count=1)
            else:
                response_data = ResponseData(data=[job_data.job_id], data_type=ResponseDataType.TEXT, total_count=1)
            return cls.success(response_data)
        except Exception as e:
            return cls.error("Error parsing response: " + str(e))



import importlib
import pkgutil
import sys
from typing import Union, Type, Dict, List, Any

from creator.base.base_app import BaseApp
from creator.base.base_request import BaseRequest


def get_creators():
    package = sys.modules['creator']  # Make sure 'src.sb_api.creator' is the correct package path
    creator_names = []
    for importer, modname, ispkg in pkgutil.walk_packages(package.__path__, package.__name__ + '.'):
        # Filter out unwanted modules
        if "creator_" not in modname or "base" in modname:
            continue

        # Try to import the module
        try:
            module = importlib.import_module(modname)
            # Find all subclasses of BaseApp in the imported module
            for name, obj in vars(module).items():
                if isinstance(obj, type) and issubclass(obj, BaseApp) and obj is not BaseApp:
                    base_app_name = obj.__name__.replace("App", "")
                    base_app_name = ''.join(['_' + i.lower() if i.isupper() else i for i in base_app_name]).lstrip('_')
                    creator_names.append(base_app_name)
        except Exception as e:
            print(f"Error importing {modname}: {e}")
    return creator_names


def get_creator_examples(openapi_output=False):
    creators = get_creators()
    examples = []
    openapi_examples = {}
    for creator in creators:
        examples.append({"name": creator, "description": f"Example for {creator} creator."})
        openapi_examples[creator] = {
            "summary": f"{creator}",
            "value": creator
        }
    return examples if not openapi_output else openapi_examples


def get_use_cases(return_creator_names: bool = False, return_param_names: bool = False) -> Union[
    Dict[str, List[BaseRequest]], Dict[BaseApp, List[BaseRequest]]]:
    package = sys.modules['creator']  # Make sure 'src.sb_api.creator' is the correct package path
    use_case_dict = {}

    for importer, modname, ispkg in pkgutil.walk_packages(package.__path__, package.__name__ + '.'):
        # Filter out unwanted modules
        if "creator_" not in modname or "base" in modname:
            continue

        # Try to import the module
        try:
            module = importlib.import_module(modname)
            # Find all subclasses of BaseApp in the imported module
            for name, obj in vars(module).items():
                if isinstance(obj, type) and issubclass(obj, BaseApp) and obj is not BaseApp:
                    params = getattr(obj, "param_classes", [])
                    if return_param_names:
                        params = [param.__name__ for param in params]
                    if not return_creator_names:
                        use_case_dict[obj] = params
                    else:
                        use_case_dict[obj.__name__] = params
        except Exception as e:
            print(f"Error importing {modname}: {e}")

    return use_case_dict


def get_use_case_examples(return_class_name=False, openapi_output=False):
    use_cases = get_use_cases(True, False)  # Always fetch class objects
    examples = []
    openapi_examples = {}
    for creator, use_case_list in use_cases.items():
        for use_case in use_case_list:
            example_param = use_case().example()
            example_execution_params = {
                "params": example_param,
                "callback_url": "http://example.com/callback"
            }
            summary = f"Execute {creator}/{use_case.__name__}_async with optional callback"
            example_name = use_case.__name__ + ("_async" if return_class_name else "")
            examples.append({"name": example_name, "value": example_execution_params})
            openapi_examples[example_name] = {
                "summary": summary,
                "value": example_execution_params
            }
    return examples if not openapi_output else openapi_examples


def get_method_examples(openapi_output=False) -> Union[Dict[str, Any], list[Dict[str, Any]]]:
    use_cases = get_use_cases(True, True)
    examples = []
    openapi_examples = {}
    for creator, use_case_list in use_cases.items():
        for method in use_case_list:
            examples.append({"name": method, "description": f"Example for {method} method."})
            openapi_examples[method] = {
                "summary": f"{method}",
                "value": method
            }
    return examples if not openapi_output else openapi_examples


def get_use_case(app_name: str, method: str) -> Union[None, Type[BaseRequest]]:
    use_cases = get_use_cases(True, False)
    method = method.replace("_async", "")
    # Convert app_name to CamelCase
    app_name = ''.join([i.capitalize() for i in app_name.split('_')])
    app_name = f"App{app_name}"
    use_case_list = use_cases.get(app_name, [])
    for use_case in use_case_list:
        if use_case.__name__.split("_")[-1] == method:
            return use_case
    return None



from setuptools import setup, find_packages

setup(
    name='gen_wrappers',
    version='0.5.9',
    author='d8ahazard',
    author_email='d8ahazard@gmail.com',
    description='A set of wrapper classes for various generative API projects',
    long_description_content_type='text/markdown',
    url='https://github.com/d8ahazard/gen_wrappers',  # Change this to your repository URL
    packages=find_packages(),
    install_requires=[
        "httpx",
        "requests",
        "pydantic",
        "fastapi"
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    python_requires='>=3.7',
)
